# Orca
#
# Copyright 2023 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-public-methods

"""Utilities for accessible states."""

from __future__ import annotations

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

from . import debug, messages
from .ax_object import AXObject


class AXUtilitiesState:
    """Utilities for accessible states."""

    @staticmethod
    def get_current_item_status_string(obj: Atspi.Accessible) -> str:
        """Returns the current item status string of obj."""

        if not AXUtilitiesState.is_active(obj):
            return ""

        result = AXObject.get_attribute(obj, "current")
        if not result:
            status = ""
        elif result == "date":
            status = messages.CURRENT_DATE
        elif result == "time":
            status = messages.CURRENT_TIME
        elif result == "location":
            status = messages.CURRENT_LOCATION
        elif result == "page":
            status = messages.CURRENT_PAGE
        elif result == "step":
            status = messages.CURRENT_STEP
        else:
            status = messages.CURRENT_ITEM
        return status

    @staticmethod
    def has_no_state(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has an empty state set"""

        if state_set is None:
            state_set = AXObject.get_state_set(obj)
        return state_set.is_empty()

    @staticmethod
    def has_popup(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the has-popup state"""

        return AXObject.has_state(obj, Atspi.StateType.HAS_POPUP, state_set)

    @staticmethod
    def has_tooltip(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the has-tooltip state"""

        return AXObject.has_state(obj, Atspi.StateType.HAS_TOOLTIP, state_set)

    @staticmethod
    def is_active(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the active state"""

        return AXObject.has_state(obj, Atspi.StateType.ACTIVE, state_set)

    @staticmethod
    def is_animated(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the animated state"""

        return AXObject.has_state(obj, Atspi.StateType.ANIMATED, state_set)

    @staticmethod
    def is_armed(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the armed state"""

        return AXObject.has_state(obj, Atspi.StateType.ARMED, state_set)

    @staticmethod
    def is_busy(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the busy state"""

        return AXObject.has_state(obj, Atspi.StateType.BUSY, state_set)

    @staticmethod
    def is_checkable(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the checkable state"""

        if state_set is None:
            state_set = AXObject.get_state_set(obj)
        if AXObject.has_state(obj, Atspi.StateType.CHECKABLE, state_set):
            return True

        if AXObject.has_state(obj, Atspi.StateType.CHECKED, state_set):
            tokens = ["AXUtilitiesState:", obj, "is checked but lacks state checkable"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        return False

    @staticmethod
    def is_checked(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the checked state"""

        if state_set is None:
            state_set = AXObject.get_state_set(obj)
        if not AXObject.has_state(obj, Atspi.StateType.CHECKED, state_set):
            return False

        if not AXObject.has_state(obj, Atspi.StateType.CHECKABLE, state_set):
            tokens = ["AXUtilitiesState:", obj, "is checked but lacks state checkable"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return True

    @staticmethod
    def is_collapsed(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the collapsed state"""

        return AXObject.has_state(obj, Atspi.StateType.COLLAPSED, state_set)

    @staticmethod
    def is_default(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the is-default state"""

        return AXObject.has_state(obj, Atspi.StateType.IS_DEFAULT, state_set)

    @staticmethod
    def is_defunct(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the defunct state"""

        return AXObject.has_state(obj, Atspi.StateType.DEFUNCT, state_set)

    @staticmethod
    def is_editable(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the editable state"""

        return AXObject.has_state(obj, Atspi.StateType.EDITABLE, state_set)

    @staticmethod
    def is_enabled(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the enabled state"""

        return AXObject.has_state(obj, Atspi.StateType.ENABLED, state_set)

    @staticmethod
    def is_expandable(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the expandable state"""

        if state_set is None:
            state_set = AXObject.get_state_set(obj)
        if AXObject.has_state(obj, Atspi.StateType.EXPANDABLE, state_set):
            return True

        if AXObject.has_state(obj, Atspi.StateType.EXPANDED, state_set):
            tokens = ["AXUtilitiesState:", obj, "is expanded but lacks state expandable"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        return False

    @staticmethod
    def is_expanded(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the expanded state"""

        if state_set is None:
            state_set = AXObject.get_state_set(obj)
        if not AXObject.has_state(obj, Atspi.StateType.EXPANDED, state_set):
            return False

        if not AXObject.has_state(obj, Atspi.StateType.EXPANDABLE, state_set):
            tokens = ["AXUtilitiesState:", obj, "is expanded but lacks state expandable"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return True

    @staticmethod
    def is_focusable(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the focusable state"""

        if state_set is None:
            state_set = AXObject.get_state_set(obj)
        if AXObject.has_state(obj, Atspi.StateType.FOCUSABLE, state_set):
            return True

        if AXObject.has_state(obj, Atspi.StateType.FOCUSED, state_set):
            tokens = ["AXUtilitiesState:", obj, "is focused but lacks state focusable"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        return False

    @staticmethod
    def is_focused(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the focused state"""

        if state_set is None:
            state_set = AXObject.get_state_set(obj)
        if not AXObject.has_state(obj, Atspi.StateType.FOCUSED, state_set):
            return False

        if not AXObject.has_state(obj, Atspi.StateType.FOCUSABLE, state_set):
            tokens = ["AXUtilitiesState:", obj, "is focused but lacks state focusable"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return True

    @staticmethod
    def is_horizontal(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the horizontal state"""

        return AXObject.has_state(obj, Atspi.StateType.HORIZONTAL, state_set)

    @staticmethod
    def is_iconified(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the iconified state"""

        return AXObject.has_state(obj, Atspi.StateType.ICONIFIED, state_set)

    @staticmethod
    def is_indeterminate(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the indeterminate state"""

        return AXObject.has_state(obj, Atspi.StateType.INDETERMINATE, state_set)

    @staticmethod
    def is_invalid_state(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the invalid_state state"""

        return AXObject.has_state(obj, Atspi.StateType.INVALID, state_set)

    @staticmethod
    def is_invalid_entry(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the invalid_entry state"""

        return AXObject.has_state(obj, Atspi.StateType.INVALID_ENTRY, state_set)

    @staticmethod
    def is_modal(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the modal state"""

        return AXObject.has_state(obj, Atspi.StateType.MODAL, state_set)

    @staticmethod
    def is_multi_line(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the multi_line state"""

        return AXObject.has_state(obj, Atspi.StateType.MULTI_LINE, state_set)

    @staticmethod
    def is_multiselectable(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the multiselectable state"""

        return AXObject.has_state(obj, Atspi.StateType.MULTISELECTABLE, state_set)

    @staticmethod
    def is_opaque(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the opaque state"""

        return AXObject.has_state(obj, Atspi.StateType.OPAQUE, state_set)

    @staticmethod
    def is_pressed(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the pressed state"""

        return AXObject.has_state(obj, Atspi.StateType.PRESSED, state_set)

    @staticmethod
    def is_read_only(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the read-only state"""

        if state_set is None:
            state_set = AXObject.get_state_set(obj)
        if AXObject.has_state(obj, Atspi.StateType.READ_ONLY, state_set):
            return True
        if AXUtilitiesState.is_editable(obj, state_set):
            return False

        # We cannot count on GTK to set the read-only state on text objects.
        return AXObject.get_role(obj) == Atspi.Role.TEXT

    @staticmethod
    def is_required(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the required state"""

        return AXObject.has_state(obj, Atspi.StateType.REQUIRED, state_set)

    @staticmethod
    def is_resizable(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the resizable state"""

        return AXObject.has_state(obj, Atspi.StateType.RESIZABLE, state_set)

    @staticmethod
    def is_selectable(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the selectable state"""

        return AXObject.has_state(obj, Atspi.StateType.SELECTABLE, state_set)

    @staticmethod
    def is_selectable_text(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the selectable-text state"""

        return AXObject.has_state(obj, Atspi.StateType.SELECTABLE_TEXT, state_set)

    @staticmethod
    def is_selected(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the selected state"""

        return AXObject.has_state(obj, Atspi.StateType.SELECTED, state_set)

    @staticmethod
    def is_sensitive(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the sensitive state"""

        return AXObject.has_state(obj, Atspi.StateType.SENSITIVE, state_set)

    @staticmethod
    def is_showing(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the showing state"""

        return AXObject.has_state(obj, Atspi.StateType.SHOWING, state_set)

    @staticmethod
    def is_showing_and_visible(
        obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None
    ) -> bool:
        """Returns true if obj has both the showing and visible states"""

        if state_set is None:
            state_set = AXObject.get_state_set(obj)
        return AXObject.has_state(obj, Atspi.StateType.SHOWING, state_set) and AXObject.has_state(
            obj, Atspi.StateType.VISIBLE, state_set
        )

    @staticmethod
    def is_single_line(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the single-line state"""

        return AXObject.has_state(obj, Atspi.StateType.SINGLE_LINE, state_set)

    @staticmethod
    def is_stale(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the stale state"""

        return AXObject.has_state(obj, Atspi.StateType.STALE, state_set)

    @staticmethod
    def is_transient(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the transient state"""

        return AXObject.has_state(obj, Atspi.StateType.TRANSIENT, state_set)

    @staticmethod
    def is_truncated(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the truncated state"""

        return AXObject.has_state(obj, Atspi.StateType.TRUNCATED, state_set)

    @staticmethod
    def is_vertical(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the vertical state"""

        return AXObject.has_state(obj, Atspi.StateType.VERTICAL, state_set)

    @staticmethod
    def is_visible(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the visible state"""

        return AXObject.has_state(obj, Atspi.StateType.VISIBLE, state_set)

    @staticmethod
    def is_visited(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the visited state"""

        return AXObject.has_state(obj, Atspi.StateType.VISITED, state_set)

    @staticmethod
    def manages_descendants(obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None) -> bool:
        """Returns true if obj has the manages-descendants state"""

        return AXObject.has_state(obj, Atspi.StateType.MANAGES_DESCENDANTS, state_set)

    @staticmethod
    def supports_autocompletion(
        obj: Atspi.Accessible, state_set: Atspi.StateSet | None = None
    ) -> bool:
        """Returns true if obj has the supports-autocompletion state"""

        return AXObject.has_state(obj, Atspi.StateType.SUPPORTS_AUTOCOMPLETION, state_set)



# Orca
#
# Copyright 2006-2008 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Provides a graphical braille display, mainly for development tasks."""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

try:
    from brlapi import KEY_CMD_ROUTE
except ImportError:
    KEY_CMD_ROUTE = None

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gdk, Gtk

from . import guilabels, script_manager

if TYPE_CHECKING:
    from collections.abc import Callable

# Attribute/Selection mask strings:
DOT_7 = "\x40"  # 01000000
DOT_8 = "\x80"  # 10000000
DOTS_78 = "\xc0"  # 11000000


# pylint: disable-next=no-member
class BrailleDot(Gtk.Alignment):
    """A single braille dot."""

    MARKUP_NORMAL = "<tt><small>%s</small></tt>"
    SYMBOL_LOWERED = "\u25cb"  # "○"
    SYMBOL_RAISED = "\u25cf"  # "●"

    def __init__(self, dot_number: int, is_raised: bool = False) -> None:
        """Create a new BrailleDot.

        Arguments:
        - dot_number: an integer reflecting the location of the dot within
          an 8-dot braille cell, using traditional braille dot values.
        - is_raised: whether the dot should initially be raised.
        """

        super().__init__()
        if dot_number in [1, 2, 3, 7]:
            self.set(1.0, 0.5, 0.0, 0.0)
            self.set_padding(3, 0, 0, 3)
        else:
            self.set(0.0, 0.5, 0.0, 0.0)
            self.set_padding(3, 0, 3, 0)

        self.label = Gtk.Label()
        self.add(self.label)
        if is_raised:
            self.raise_dot()
        else:
            self.lower_dot()

    def raise_dot(self) -> None:
        """Raise the dot (make it visible/filled)."""
        self.set(0.5, 0.5, 0, 0)
        self.label.set_markup(self.MARKUP_NORMAL % self.SYMBOL_RAISED)

    def lower_dot(self) -> None:
        """Lower the dot (make it invisible/empty)."""
        self.set(0.5, 0.5, 0, 0)
        self.label.set_markup(self.MARKUP_NORMAL % self.SYMBOL_LOWERED)


class BrailleCell(Gtk.Button):
    """A single graphical braille cell with cursor routing capability."""

    MARKUP_NORMAL = "<tt><span size='x-large'><b>%s</b></span></tt>"
    MARKUP_BRAILLE = "<tt><span size='xx-large'><b>%s</b></span></tt>"
    MARKUP_CURSOR_CELL = "<u>%s</u>"

    def __init__(self, position: int) -> None:
        """Create a new BrailleCell.

        Arguments:
        - position: The location of the cell with respect to the monitor.
        """

        super().__init__()
        self.get_style_context().add_class("braille-cell")
        self.set_size_request(36, 52)
        self._position = position
        self._displayed_char = Gtk.Label()
        self._displayed_char.set_valign(Gtk.Align.END)
        self._dot7 = BrailleDot(7)
        self._dot8 = BrailleDot(8)

        grid = Gtk.Grid()
        grid.attach(self._displayed_char, 0, 0, 2, 3)
        grid.attach(self._dot7, 0, 3, 1, 1)
        grid.attach(self._dot8, 1, 3, 1, 1)
        self.add(grid)  # pylint: disable=no-member

        self.connect("clicked", self._on_cell_clicked)

    def _on_cell_clicked(self, _widget: Gtk.Button) -> None:
        """Callback for the 'clicked' signal on the push button."""

        if KEY_CMD_ROUTE is None:
            return

        script = script_manager.get_manager().get_active_script()
        if script is None:
            return

        # pylint: disable=import-outside-toplevel
        from . import braille_presenter, input_event

        fake_key_press = {}
        fake_key_press["command"] = KEY_CMD_ROUTE
        fake_key_press["argument"] = self._position
        event = input_event.BrailleEvent(fake_key_press)
        braille_presenter.get_presenter().process_routing_key(script, event)

    def clear(self) -> None:
        """Clears the braille cell."""

        self._displayed_char.set_markup(self.MARKUP_NORMAL % " ")
        self._dot7.lower_dot()
        self._dot8.lower_dot()

    def display(self, char: str, mask: str | None = None, is_cursor_cell: bool = False) -> None:
        """Displays the specified character in the cell.

        Arguments:
        - char: The character to display in the cell.
        - mask: Optional mask for displaying attributes.
        - is_cursor_cell: If True, the cursor/caret is at this cell.
        """

        if char == "&":
            char = "&amp;"
        elif char == "<":
            char = "&lt;"
        elif char == "\t":
            char = "$t"

        is_braille = len(char) == 1 and 0x2800 <= ord(char) <= 0x28FF
        markup = self.MARKUP_BRAILLE if is_braille else self.MARKUP_NORMAL
        if is_cursor_cell:
            markup = markup % self.MARKUP_CURSOR_CELL
        self._displayed_char.set_markup(markup % char)

        if mask in [DOT_7, DOTS_78]:
            self._dot7.raise_dot()
        if mask in [DOT_8, DOTS_78]:
            self._dot8.raise_dot()


class BrailleMonitor(Gtk.Window):
    """Displays a GUI braille monitor that mirrors what would be displayed
    by Orca on a connected, configured, and enabled braille display."""

    _shared_css_provider: Gtk.CssProvider | None = None

    def __init__(
        self,
        num_cells: int = 32,
        on_close: Callable[[], None] | None = None,
        foreground: str = "#000000",
        background: str = "#ffffff",
    ) -> None:
        """Create a new BrailleMonitor."""

        # pylint: disable=no-member

        super().__init__()
        self._on_close = on_close
        self._foreground = foreground
        self._background = background
        self.set_title(guilabels.BRAILLE_MONITOR)
        self.set_icon_name("orca")

        titlebar = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        title_label = Gtk.Label(label=guilabels.BRAILLE_MONITOR)
        titlebar.set_center_widget(title_label)
        close_btn = Gtk.Button.new_from_icon_name(
            "window-close-symbolic",
            Gtk.IconSize.LARGE_TOOLBAR,
        )
        close_btn.set_relief(Gtk.ReliefStyle.NONE)
        close_btn.connect("clicked", self._on_close_clicked)
        titlebar.pack_end(close_btn, False, False, 0)
        self.set_titlebar(titlebar)

        self.get_style_context().add_class("braille-monitor")
        titlebar.get_style_context().add_class("braille-monitor-titlebar")
        title_label.get_style_context().add_class("braille-monitor-title")
        close_btn.get_style_context().add_class("braille-monitor-close")

        cell_box = Gtk.Grid()
        self.add(cell_box)
        if os.environ.get("XDG_SESSION_TYPE", "").lower() == "wayland":
            cell_box.set_sensitive(False)

        self.cells: list[BrailleCell] = []
        for i in range(num_cells):
            cell = BrailleCell(i)
            cell_box.attach(cell, i, 0, 1, 1)
            self.cells.append(cell)

        if BrailleMonitor._shared_css_provider is None:
            BrailleMonitor._shared_css_provider = Gtk.CssProvider()
            Gtk.StyleContext.add_provider_for_screen(
                Gdk.Screen.get_default(),
                BrailleMonitor._shared_css_provider,
                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
            )
        self._apply_css()

        self.set_resizable(False)
        self.set_keep_above(True)
        self.set_property("accept-focus", False)
        self.set_skip_taskbar_hint(True)
        self.set_skip_pager_hint(True)

        self.connect("delete-event", self._on_delete_event)

    def reapply_css(
        self,
        foreground: str | None = None,
        background: str | None = None,
    ) -> None:
        """Reapplies CSS styling (e.g. after color changes)."""

        if foreground is not None:
            self._foreground = foreground
        if background is not None:
            self._background = background
        self._apply_css()

    def _apply_css(self) -> None:
        """Apply CSS styling for colors and minimal titlebar."""

        bg = self._background
        fg = self._foreground
        css = (
            f".braille-monitor {{ background-color: {bg}; }}\n"
            f".braille-monitor-titlebar {{ min-height: 0; padding: 0; margin: 0; "
            f"background-color: {bg}; border: none; box-shadow: none; }}\n"
            f".braille-monitor-title {{ color: {fg}; font-size: small; font-weight: bold; }}\n"
            f".braille-monitor-close {{ min-height: 16px; min-width: 16px; "
            f"padding: 2px; margin: 0; color: {fg}; }}\n"
            f".braille-monitor-close:hover {{ opacity: 0.7; }}\n"
            f".braille-monitor .braille-cell {{ "
            f"background: {bg}; color: {fg}; border-color: {fg}; }}\n"
            f".braille-monitor .braille-cell label {{ color: {fg}; }}"
        )
        if BrailleMonitor._shared_css_provider is not None:
            BrailleMonitor._shared_css_provider.load_from_data(css.encode())

    def _on_close_clicked(self, _button: Gtk.Button) -> None:
        """Handle the close button click."""

        if self._on_close is not None:
            self._on_close()

    def _on_delete_event(self, _window: Gtk.Window, _event: Gdk.Event) -> bool:
        """Intercept the close request to toggle the monitor off."""

        if self._on_close is not None:
            self._on_close()
        return True

    def clear(self) -> None:
        """Clears the braille monitor display."""

        for cell in self.cells:
            cell.clear()

    def write_text(self, cursor_cell: int, string: str, mask: str | None = None) -> None:
        """Display the given text and highlight the given cursor cell.

        Arguments:
        - cursor_cell: 1-based index of cell with cursor (0 means no cursor)
        - string: text to display (length must be <= num cells)
        - mask: optional attribute mask
        """

        self.clear()
        length = min(len(string), len(self.cells))
        for i in range(length):
            is_cursor_cell = i == cursor_cell - 1
            try:
                cell_mask = mask[i] if mask else None
            except (IndexError, TypeError):
                cell_mask = None
            self.cells[i].display(string[i], cell_mask, is_cursor_cell)



# Orca
#
# Copyright 2023-2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Utilities for obtaining selection-related information about accessible objects."""

from __future__ import annotations

from typing import TYPE_CHECKING

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

from . import ax_cache_manager, debug
from .ax_object import AXObject
from .ax_selection import AXSelection
from .ax_table import AXTable
from .ax_utilities_collection import AXUtilitiesCollection
from .ax_utilities_object import AXUtilitiesObject
from .ax_utilities_role import AXUtilitiesRole
from .ax_utilities_state import AXUtilitiesState
from .ax_utilities_table import AXUtilitiesTable

if TYPE_CHECKING:
    from collections.abc import Callable


class _AXUtilitiesSelectionCache:
    """Provides selection-specific access to manager-backed cached values."""

    IS_ALL_ITEMS_SELECTED = "AXUtilitiesSelection.is-all-items-selected"

    def __init__(self) -> None:
        self._manager = ax_cache_manager.get_manager()
        self._manager.register_cache(
            self,
            self.IS_ALL_ITEMS_SELECTED,
            lifetime=ax_cache_manager.Lifetime.PROCESS,
            clear_interval_seconds=None,
        )
        self._all_items_selected_cache = self._manager.get_cache(self, self.IS_ALL_ITEMS_SELECTED)

    def is_all_items_selected(self, obj: Atspi.Accessible) -> bool:
        """Returns the cached all-items-selected state for obj."""

        if self._all_items_selected_cache is None:
            return False

        return self._all_items_selected_cache.get(ax_cache_manager.get_object_key(obj), False)

    def set_all_items_selected(self, obj: Atspi.Accessible, selected: bool) -> None:
        """Stores the all-items-selected state for obj."""

        if self._all_items_selected_cache is not None:
            self._all_items_selected_cache.put(ax_cache_manager.get_object_key(obj), selected)


class AXUtilitiesSelection:
    """Utilities for obtaining selection-related information about accessible objects."""

    _CACHE = _AXUtilitiesSelectionCache()

    @staticmethod
    def get_all_items_selected_state(obj: Atspi.Accessible) -> bool:
        """Returns the cached all-items-selected state for obj."""

        return AXUtilitiesSelection._CACHE.is_all_items_selected(obj)

    @staticmethod
    def set_all_items_selected_state(obj: Atspi.Accessible, selected: bool) -> None:
        """Sets the cached all-items-selected state for obj."""

        AXUtilitiesSelection._CACHE.set_all_items_selected(obj, selected)

    @staticmethod
    def get_selected_children(obj: Atspi.Accessible) -> list[Atspi.Accessible]:
        """Returns a list of all the selected children of obj."""

        if obj is None:
            return []

        count = AXSelection.get_selected_child_count(obj)
        if not count and AXUtilitiesRole.is_combo_box(obj):
            if AXObject.supports_collection(obj):
                container = AXUtilitiesCollection.find_first_with_role(
                    obj, [Atspi.Role.MENU, Atspi.Role.LIST_BOX]
                )
            else:
                container = AXUtilitiesObject.find_descendant(
                    obj,
                    lambda x: AXUtilitiesRole.is_menu(x) or AXUtilitiesRole.is_list_box(x),
                )
            return AXUtilitiesSelection.get_selected_children(container)

        children = set()
        for i in range(count):
            child = AXSelection.get_selected_child(obj, i)
            if child is not None:
                children.add(child)

        if obj in children:
            tokens = ["AXUtilitiesSelection:", obj, "claims to be its own selected child"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            children.remove(obj)

        result = list(children)
        if len(result) != count:
            tokens = ["AXUtilitiesSelection: Selected child count of", obj, f"is {count}"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return result

    @staticmethod
    def selected_children(obj: Atspi.Accessible) -> list[Atspi.Accessible]:
        """Returns a list of selected children in obj."""

        if AXUtilitiesTable.is_spreadsheet_table(obj):
            return []

        return AXUtilitiesSelection.get_selected_children(obj)

    @staticmethod
    def get_selection_container(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the selection container for obj."""

        if not obj:
            return None

        if AXUtilitiesRole.is_paragraph(obj) or AXUtilitiesState.is_editable(obj):
            return None

        if AXObject.supports_selection(obj):
            return obj

        rolemap = {
            Atspi.Role.CANVAS: [Atspi.Role.LAYERED_PANE],
            Atspi.Role.ICON: [Atspi.Role.LAYERED_PANE],
            Atspi.Role.LIST_ITEM: [Atspi.Role.LIST_BOX],
            Atspi.Role.TREE_ITEM: [Atspi.Role.TREE, Atspi.Role.TREE_TABLE],
            Atspi.Role.TABLE_CELL: [Atspi.Role.TABLE, Atspi.Role.TREE_TABLE],
            Atspi.Role.TABLE_ROW: [Atspi.Role.TABLE, Atspi.Role.TREE_TABLE],
        }

        matching_roles = rolemap.get(AXObject.get_role(obj))

        def is_match(x: Atspi.Accessible) -> bool:
            if matching_roles and AXObject.get_role(x) not in matching_roles:
                return False
            return AXObject.supports_selection(x)

        return AXUtilitiesObject.find_ancestor(obj, is_match)

    @staticmethod
    def selectable_child_count(obj: Atspi.Accessible) -> int:
        """Returns the number of selectable children in obj."""

        if not AXObject.supports_selection(obj):
            return 0

        if AXObject.supports_table(obj):
            rows = AXTable.get_row_count(obj)
            return max(0, rows)

        rolemap = {
            Atspi.Role.LIST_BOX: [Atspi.Role.LIST_ITEM],
            Atspi.Role.TREE: [Atspi.Role.TREE_ITEM],
        }

        role = AXObject.get_role(obj)
        if role not in rolemap:
            return AXObject.get_child_count(obj)

        child_roles = rolemap[role]
        if AXObject.supports_collection(obj):
            return len(AXUtilitiesCollection.find_all_with_role(obj, child_roles))

        def is_match(x: Atspi.Accessible) -> bool:
            return AXObject.get_role(x) in child_roles

        return len(AXUtilitiesObject.find_all_descendants(obj, is_match))

    @staticmethod
    def selected_child_count(obj: Atspi.Accessible) -> int:
        """Returns the number of selected children in obj."""

        if AXObject.supports_table(obj):
            return AXTable.get_selected_row_count(obj)
        return AXSelection.get_selected_child_count(obj)

    @staticmethod
    def all_items_selected(obj: Atspi.Accessible) -> bool:
        """Returns True if all items in obj are selected."""

        if not AXObject.supports_selection(obj):
            return False

        if AXUtilitiesState.is_expandable(obj) and not AXUtilitiesState.is_expanded(obj):
            return False

        if AXUtilitiesRole.is_combo_box(obj) or AXUtilitiesRole.is_menu(obj):
            return False

        child_count = AXObject.get_child_count(obj)
        if child_count == AXSelection.get_selected_child_count(obj):
            child = AXSelection.get_selected_child(obj, 0)
            if AXObject.get_parent(child) != obj:
                return False

            tokens = ["AXUtilitiesSelection: All", child_count, "children believed to be selected"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        return AXUtilitiesTable.all_cells_are_selected(obj)

    @staticmethod
    def get_selected_child_for_focus(
        obj: Atspi.Accessible,
        focus: Atspi.Accessible | None,
        should_skip: Callable[[Atspi.Accessible], bool] | None = None,
    ) -> Atspi.Accessible | None:
        """Returns the selected child of obj that should become the locus of focus, or None."""

        selected = AXUtilitiesSelection.selected_children(obj)
        if focus in selected:
            tokens = ["AXUtilitiesSelection: Focus", focus, "is already in selected children"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        for child in selected:
            if should_skip is not None and should_skip(child):
                continue

            if AXUtilitiesObject.is_ancestor(focus, child):
                tokens = ["AXUtilitiesSelection: Child", child, "is ancestor of focus"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return None

            if (
                AXUtilitiesRole.is_page_tab(child)
                and focus is not None
                and AXObject.get_name(child) == AXObject.get_name(focus)
                and not AXUtilitiesState.is_focused(obj)
            ):
                tokens = ["AXUtilitiesSelection: Selected", child, "matches focus", focus]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return None

            return child

        return None



"""Orca Screen Reader"""

from __future__ import annotations

import os

_ORCA_DIR = os.path.dirname(os.path.realpath(__file__))


def is_orca(filename: str, module_prefix: str = "") -> bool:
    """Return True if filename is Orca code, optionally matching module_prefix."""

    filepath = os.path.realpath(filename) if filename else ""
    module_name = os.path.splitext(os.path.basename(filepath))[0] if filepath else ""
    try:
        is_orca_code = os.path.commonpath([_ORCA_DIR, filepath]) == _ORCA_DIR
    except ValueError:
        return False

    return is_orca_code and (not module_prefix or module_name.startswith(module_prefix))



# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Provides interactive math navigation via MathCAT."""

from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING

from . import (
    braille,
    clipboard,
    command_manager,
    dbus_service,
    debug,
    document_presenter,
    focus_manager,
    guilabels,
    input_event,
    math_navigator_command_definitions,
    math_presenter,
    messages,
    presentation_manager,
    script_manager,
    speech_presenter,
)
from .ax_utilities_math import AXUtilitiesMath
from .ax_utilities_text import CaretSetReason
from .extension import Extension

try:
    from . import libmathcat_py  # type: ignore[attr-defined]
except ImportError:
    libmathcat_py = None  # type: ignore[assignment]

if TYPE_CHECKING:
    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

    from .scripts import default


class MathNavigator(Extension):
    """Provides interactive exploration of math expressions via MathCAT."""

    GROUP_LABEL = guilabels.KB_GROUP_MATH_NAVIGATION

    def __init__(self) -> None:
        self._active: bool = False
        self._math_object: Atspi.Accessible | None = None
        self._last_command: str = ""
        self._last_nav_id: tuple[str, int] = ("", 0)
        self._suspended: bool = True
        super().__init__()

    @staticmethod
    def navigation_command(func):
        """Decorator that logs the command, then dispatches to it."""

        @wraps(func)
        def wrapper(self, script, event=None, notify_user=True) -> bool:
            tokens = [
                "MATH NAVIGATOR:",
                func,
                "\nScript:",
                script,
                "\nEvent:",
                event,
                "\nnotify_user:",
                notify_user,
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return func(self, script, notify_user)

        return wrapper

    @dbus_service.getter
    def is_active(self) -> bool:
        """Returns True if currently navigating a math expression."""

        return self._active

    @dbus_service.getter
    def get_supported_commands(self) -> list[str]:
        """Returns the MathCAT navigation commands supported by this navigator."""

        return math_navigator_command_definitions.get_supported_commands()

    def _register_commands(self) -> None:
        if libmathcat_py is None:
            msg = f"EXTENSION: {self.module_name} MathCAT not available. Skipping command setup."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        manager = command_manager.get_manager()
        for command in math_navigator_command_definitions.get_commands(self):
            manager.add_command(command)
        manager.set_group_suspended(self.GROUP_LABEL, True)

        msg = f"EXTENSION: {self.module_name} Commands set up."
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _is_active_script(self, script):
        active_script = script_manager.get_manager().get_active_script()
        if active_script == script:
            return True

        tokens = ["MATH NAVIGATOR:", script, "is not the active script", active_script]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return False

    def suspend_commands(
        self,
        script: default.Script,
        suspended: bool,
        reason: str = "",
    ) -> None:
        """Suspends math navigation independent of the enabled setting."""

        if not (script and self._is_active_script(script)):
            return

        msg = f"MATH NAVIGATOR: Commands suspended: {suspended}"
        if reason:
            msg += f": {reason}"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        self._suspended = suspended
        command_manager.get_manager().set_group_suspended(
            guilabels.KB_GROUP_MATH_NAVIGATION,
            suspended,
        )

    def enter_math_mode(self, obj: Atspi.Accessible) -> bool:
        """Enters math navigation mode for the expression containing obj."""

        already_active = self._active
        # Always (re-)enter so pressing the command again restarts navigation at the outermost
        # element rather than leaving the caret wherever it last was in the expression.
        if not self._enter(obj):
            return False

        if not already_active:
            # Navigation already read the expression, so first entry only confirms the mode.
            presentation_manager.get_manager().present_message(messages.MATH_NAVIGATION_ENTERED)
            return True

        # Re-pressing restarts at the outermost element and re-presents the whole expression.
        math_obj = self._math_object or obj
        if speech := math_presenter.get_presenter().get_speech_for_math(math_obj):
            presentation_manager.get_manager().speak_accessible_text(math_obj, speech)
        return True

    @dbus_service.command
    @navigation_command
    def enter_math_mode_command(self, script: default.Script, notify_user: bool = True) -> bool:
        """Enters math navigation mode if the current focus is on math."""

        obj = focus_manager.get_manager().get_locus_of_focus()
        if not (math_root := AXUtilitiesMath.find_math_root(obj)):
            if notify_user:
                presentation_manager.get_manager().present_message(
                    messages.MATH_NAVIGATION_NOT_IN_MATH,
                )
            return True

        document_presenter.get_presenter().enter_math_navigation(script, math_root)
        return True

    @dbus_service.command
    @navigation_command
    def exit_math_mode(self, script: default.Script, notify_user: bool = True) -> bool:
        """Exits math navigation mode."""

        if not self._active:
            return False

        self._do_command("Exit")
        self.reset(script)
        document_presenter.get_presenter().exit_math_navigation(script)

        if notify_user:
            presentation_manager.get_manager().present_message(
                messages.MATH_NAVIGATION_EXITED,
            )
        return True

    @dbus_service.command
    def copy_to_clipboard(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Copies the current math navigation node to the clipboard."""

        tokens = [
            "MATH NAVIGATOR: copy_to_clipboard. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not self._active:
            return False

        content = math_presenter.get_presenter().get_navigation_content_for_copy()
        if not content:
            return False

        clipboard.get_presenter().set_text(content)
        return True

    @dbus_service.parameterized_command
    def execute_mathcat_command(
        self,
        mathcat_command: str,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Executes a MathCAT navigation command. Use get_supported_commands for valid values."""

        if script is None:
            return False
        return self._execute_command(mathcat_command, script, event, notify_user)

    def _execute_command(
        self,
        mathcat_command: str,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Executes a MathCAT navigation command and presents the result."""

        tokens = [
            f"MATH NAVIGATOR: _execute_command: {mathcat_command}. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not self._active:
            return False

        at_boundary = (
            mathcat_command in self._MOVEMENT_COMMANDS
            and mathcat_command == self._last_command
            and self._last_nav_id == self._get_nav_location()
        )

        speech = self._do_command(mathcat_command)
        self._last_command = mathcat_command
        self._last_nav_id = self._get_nav_location()

        if notify_user:
            self._present(speech, mathcat_command)
            if at_boundary and speech_presenter.get_presenter().use_verbose_speech():
                presentation_manager.get_manager().present_message(
                    messages.MATH_NAVIGATION_ESCAPE_HINT,
                )
        return True

    def _enter(self, math_root: Atspi.Accessible) -> bool:
        """Enters math navigation for the given math root element."""

        mathml = AXUtilitiesMath.get_mathml(math_root)
        if not mathml:
            return False

        if not math_presenter.get_presenter().is_available():
            return False

        try:
            libmathcat_py.SetMathML(mathml)
        except OSError as err:
            msg = f"MATH NAVIGATOR: SetMathML failed: {err}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        self._math_object = math_root
        self._active = True
        self._last_command = ""
        self._last_nav_id = ("", 0)

        focus_manager.get_manager().emit_region_changed(
            math_root,
            mode=focus_manager.MATH_NAVIGATOR,
        )

        msg = "MATH NAVIGATOR: Entered math navigation."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    def reset(self, script: default.Script) -> None:
        """Resets navigator state and positions the caret on the math object."""

        math_obj = self._math_object
        self._active = False
        self._math_object = None
        self._last_command = ""
        self._last_nav_id = ("", 0)

        if math_obj is not None:
            script.utilities.set_caret_position(math_obj, 0, reason=CaretSetReason.MATH_NAVIGATION)

        msg = "MATH NAVIGATOR: Exited math navigation."
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _get_nav_location(self) -> tuple[str, int]:
        """Returns the current navigation location as (id, offset)."""

        try:
            return libmathcat_py.GetNavigationMathMLId()
        except OSError:
            return ("", 0)

    def _do_command(self, command: str) -> str:
        """Sends a navigation command to MathCAT and returns the speech result."""

        try:
            speech = libmathcat_py.DoNavigateCommand(command)
        except OSError as err:
            msg = f"MATH NAVIGATOR: DoNavigateCommand({command}) failed: {err}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return ""

        msg = f"MATH NAVIGATOR: {command} -> {speech}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return speech

    _MOVEMENT_COMMANDS = frozenset(
        (
            "MoveNext",
            "MovePrevious",
            "MoveStart",
            "MoveEnd",
            "MoveLineStart",
            "MoveLineEnd",
            "MoveColumnStart",
            "MoveColumnEnd",
            "MoveCellPrevious",
            "MoveCellNext",
            "MoveCellUp",
            "MoveCellDown",
            "MoveLastLocation",
            "ZoomIn",
            "ZoomOut",
            "ZoomInAll",
            "ZoomOutAll",
        )
    )

    _INFORMATIONAL_COMMANDS = frozenset(
        (
            "WhereAmI",
            "WhereAmIAll",
            "ToggleZoomLockUp",
            "ToggleZoomLockDown",
            "ToggleSpeakMode",
        )
    )

    def _present(self, speech: str, command: str = "") -> None:
        """Presents the navigation result via speech and braille."""

        if speech:
            if command in self._INFORMATIONAL_COMMANDS:
                presentation_manager.get_manager().present_message(speech)
            else:
                presentation_manager.get_manager().speak_accessible_text(
                    self._math_object,
                    speech,
                )

        try:
            braille_string = libmathcat_py.GetNavigationBraille()
        except OSError:
            braille_string = ""

        if braille_string and self._math_object is not None:
            region = braille.Component(
                self._math_object,
                braille_string,
                contracted=False,
            )
            line = braille.Line(region)
            braille.display_line(line, region)


_navigator: MathNavigator = MathNavigator()


def get_navigator() -> MathNavigator:
    """Returns the Math Navigator"""

    return _navigator



# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Command definitions for document presentation."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import cmdnames, keybindings
from .command import Command, KeyboardCommand

if TYPE_CHECKING:
    from .document_presenter import DocumentPresenter


def get_commands(owner: DocumentPresenter) -> list[Command]:
    """Returns commands for document presentation."""

    kb_orca_a = keybindings.KeyBinding("a", keybindings.ORCA_MODIFIER_MASK)
    kb_orca_a_2 = keybindings.KeyBinding("a", keybindings.ORCA_MODIFIER_MASK, click_count=2)
    kb_orca_a_3 = keybindings.KeyBinding("a", keybindings.ORCA_MODIFIER_MASK, click_count=3)

    return [
        KeyboardCommand(
            "toggle_presentation_mode",
            owner.toggle_presentation_mode,
            owner.GROUP_LABEL,
            cmdnames.TOGGLE_PRESENTATION_MODE,
            desktop_keybinding=kb_orca_a,
            laptop_keybinding=kb_orca_a,
        ),
        KeyboardCommand(
            "enable_sticky_focus_mode",
            owner.enable_sticky_focus_mode,
            owner.GROUP_LABEL,
            cmdnames.SET_FOCUS_MODE_STICKY,
            desktop_keybinding=kb_orca_a_2,
            laptop_keybinding=kb_orca_a_2,
        ),
        KeyboardCommand(
            "enable_sticky_browse_mode",
            owner.enable_sticky_browse_mode,
            owner.GROUP_LABEL,
            cmdnames.SET_BROWSE_MODE_STICKY,
            desktop_keybinding=kb_orca_a_3,
            laptop_keybinding=kb_orca_a_3,
        ),
    ]



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Command definitions for speech management."""

from __future__ import annotations

import functools
from typing import TYPE_CHECKING

from . import cmdnames, keybindings
from .command import Command, KeyboardCommand

if TYPE_CHECKING:
    from .speech_manager import SpeechManager


def get_commands(owner: SpeechManager) -> list[Command]:
    """Returns commands for speech management."""

    kb_orca_s = keybindings.KeyBinding("s", keybindings.ORCA_MODIFIER_MASK)

    return [
        KeyboardCommand(
            "cycleCapitalizationStyleHandler",
            owner.cycle_capitalization_style,
            owner.GROUP_LABEL,
            cmdnames.CYCLE_CAPITALIZATION_STYLE,
        ),
        KeyboardCommand(
            "cycleSpeakingPunctuationLevelHandler",
            owner.cycle_punctuation_level,
            owner.GROUP_LABEL,
            cmdnames.CYCLE_PUNCTUATION_LEVEL,
        ),
        KeyboardCommand(
            "cycleSynthesizerHandler",
            owner.cycle_synthesizer,
            owner.GROUP_LABEL,
            cmdnames.CYCLE_SYNTHESIZER,
        ),
        KeyboardCommand(
            "cycleVoiceSetHandler",
            owner.cycle_voice_set,
            owner.GROUP_LABEL,
            cmdnames.CYCLE_VOICE_SET,
        ),
        KeyboardCommand(
            "toggleSilenceSpeechHandler",
            owner.toggle_speech,
            owner.GROUP_LABEL,
            cmdnames.TOGGLE_SPEECH,
            desktop_keybinding=kb_orca_s,
            laptop_keybinding=kb_orca_s,
        ),
        KeyboardCommand(
            "decreaseSpeechRateHandler",
            owner.decrease_rate,
            owner.GROUP_LABEL,
            cmdnames.DECREASE_SPEECH_RATE,
        ),
        KeyboardCommand(
            "increaseSpeechRateHandler",
            owner.increase_rate,
            owner.GROUP_LABEL,
            cmdnames.INCREASE_SPEECH_RATE,
        ),
        KeyboardCommand(
            "decreaseSpeechPitchHandler",
            owner.decrease_pitch,
            owner.GROUP_LABEL,
            cmdnames.DECREASE_SPEECH_PITCH,
        ),
        KeyboardCommand(
            "increaseSpeechPitchHandler",
            owner.increase_pitch,
            owner.GROUP_LABEL,
            cmdnames.INCREASE_SPEECH_PITCH,
        ),
        KeyboardCommand(
            "decreaseSpeechInflectionHandler",
            owner.decrease_pitch_range,
            owner.GROUP_LABEL,
            cmdnames.DECREASE_SPEECH_INFLECTION,
        ),
        KeyboardCommand(
            "increaseSpeechInflectionHandler",
            owner.increase_pitch_range,
            owner.GROUP_LABEL,
            cmdnames.INCREASE_SPEECH_INFLECTION,
        ),
        KeyboardCommand(
            "decreaseSpeechVolumeHandler",
            owner.decrease_volume,
            owner.GROUP_LABEL,
            cmdnames.DECREASE_SPEECH_VOLUME,
        ),
        KeyboardCommand(
            "increaseSpeechVolumeHandler",
            owner.increase_volume,
            owner.GROUP_LABEL,
            cmdnames.INCREASE_SPEECH_VOLUME,
        ),
    ]


def get_voice_set_commands(owner: SpeechManager) -> list[KeyboardCommand]:
    """Returns one unbound activation command per available voice set."""

    return [
        KeyboardCommand(
            f"switch-voice-set-{set_id}",
            functools.partial(owner.activate_voice_set, set_id),
            owner.GROUP_LABEL,
            cmdnames.SWITCH_VOICE_SET % owner._voice_set_display_name(set_id),  # pylint: disable=protected-access
            desktop_keybinding=None,
            laptop_keybinding=None,
            transient=True,
        )
        for set_id in owner.get_available_voice_sets()
    ]



# Orca
#
# Copyright 2024-2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Module for sleep mode"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from . import (
    dbus_service,
    debug,
    gsettings_registry,
    guilabels,
    input_event,
    messages,
    presentation_manager,
    script_manager,
    sleep_mode_manager_command_definitions,
)
from .ax_object import AXObject
from .ax_utilities import AXUtilities
from .extension import Extension

if TYPE_CHECKING:
    from gi.repository import Atspi

    from .command import Command
    from .scripts import default
    from .sleep_mode_manager_preferences_grid import SleepModePreferencesGrid


@gsettings_registry.get_registry().gsettings_schema(
    "org.gnome.Orca.SleepMode",
    name="sleep-mode",
)
class SleepModeManager(Extension):
    """Provides sleep mode implementation."""

    GROUP_LABEL = guilabels.KB_GROUP_SLEEP_MODE
    COMMAND_NAME = "toggle_sleep_mode"
    _SCHEMA = "sleep-mode"

    def __init__(self) -> None:
        self._apps: list[int] = []
        self._runtime_exceptions: set[str] = set()
        super().__init__()

    def _get_commands(self) -> list[Command]:
        """Returns commands for registration."""

        return sleep_mode_manager_command_definitions.get_commands(self)

    def _is_on_persistent_list(self, app: Atspi.Accessible) -> bool:
        """Returns True if the app is on the persistent sleep mode list."""

        app_name = AXObject.get_name(app)
        if not app_name:
            return False
        return app_name in self.get_sleep_mode_apps()

    def is_active_for_app(self, app: Atspi.Accessible) -> bool:
        """Returns True if sleep mode is active for app."""

        if not app:
            return False

        app_name = AXObject.get_name(app)
        if app_name in self._runtime_exceptions:
            tokens = ["SLEEP MODE MANAGER: Skipping (runtime exception for", app, ")"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        if hash(app) in self._apps:
            tokens = ["SLEEP MODE MANAGER: Is active for", app, "(runtime toggle)"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        if self._is_on_persistent_list(app):
            tokens = ["SLEEP MODE MANAGER: Is active for", app, "(persistent setting)"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        return False

    def on_app_deactivated(self, app: Atspi.Accessible) -> None:
        """Clears any runtime exception so persistent sleep re-engages on return."""

        app_name = AXObject.get_name(app)
        if app_name and app_name in self._runtime_exceptions:
            self._runtime_exceptions.discard(app_name)
            tokens = ["SLEEP MODE MANAGER: Cleared runtime exception for", app]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

    @gsettings_registry.get_registry().gsetting(
        key="apps",
        schema="sleep-mode",
        gtype="as",
        default=[],
        summary="Applications that automatically use sleep mode",
    )
    @dbus_service.getter
    def get_sleep_mode_apps(self) -> list[str]:
        """Returns the list of apps that should automatically use sleep mode."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            "apps",
            "as",
            default=[],
        )

    @dbus_service.setter
    def set_sleep_mode_apps(self, apps: list[str]) -> bool:
        """Sets the list of apps that should automatically use sleep mode."""

        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            "apps",
            apps,
        )
        return True

    def get_running_app_names(self) -> list[str]:
        """Returns sorted names of running applications with windows, excluding Orca."""

        orca_pid = os.getpid()
        names: set[str] = set()
        for app in AXUtilities.get_all_applications(must_have_window=True):
            if AXUtilities.get_process_id(app) == orca_pid:
                continue
            name = AXObject.get_name(app)
            if name:
                names.add(name)
        return sorted(names)

    def create_preferences_grid(self) -> SleepModePreferencesGrid:
        """Returns the preferences grid for sleep mode settings."""

        # pylint: disable-next=import-outside-toplevel
        from .sleep_mode_manager_preferences_grid import SleepModePreferencesGrid

        return SleepModePreferencesGrid(self)

    @dbus_service.command
    def toggle_sleep_mode(
        self,
        script: default.Script | None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Toggles sleep mode for the active application."""

        tokens = [
            "SLEEP MODE MANAGER: toggle_sleep_mode. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not (script and script.app):
            return True

        _script_manager = script_manager.get_manager()
        if self.is_active_for_app(script.app):
            app_name = AXObject.get_name(script.app)
            if hash(script.app) in self._apps:
                self._apps.remove(hash(script.app))
            if self._is_on_persistent_list(script.app) and app_name:
                self._runtime_exceptions.add(app_name)
            new_script = _script_manager.get_script(script.app)
            if notify_user:
                presentation_manager.get_manager().present_message(
                    messages.SLEEP_MODE_DISABLED_FOR % app_name,
                )
            _script_manager.set_active_script(new_script, "Sleep mode toggled off")
            return True

        if notify_user:
            msg = messages.SLEEP_MODE_ENABLED_FOR % AXObject.get_name(script.app)
            manager = presentation_manager.get_manager()
            manager.speak_message(msg)

            # Don't restore previous braille content because Orca is no longer active.
            manager.present_braille_message(msg, restore_previous=False)
        _script_manager.set_active_script(
            _script_manager.get_or_create_sleep_mode_script(script.app),  # type: ignore[arg-type]
            "Sleep mode toggled on",
        )
        self._apps.append(hash(script.app))
        return True


_manager: SleepModeManager = SleepModeManager()


def get_manager() -> SleepModeManager:
    """Returns the Sleep Mode Manager singleton."""
    return _manager
